{-# LANGUAGE ExplicitNamespaces, FlexibleInstances, MultiParamTypeClasses, ScopedTypeVariables, TypeOperators, UndecidableInstances #-}
module GEODE.Metadata.Types
  ( Has(..)
  , ToJSONObject(..)
  , type (@)(..) ) where

import Data.Aeson (FromJSON(..), Object, Series, ToJSON(..), Value(..), pairs)
import Data.Aeson.KeyMap as Object (union)
import Data.Csv (DefaultOrdered(..), FromNamedRecord(..), ToNamedRecord(..))
import Data.HashMap.Strict as Hash (union)

infixl 9 @
infixl 9 :@:
data a @ b = a :@: b

class Has a b where
  get :: b -> a

instance Has a a where
  get = id

instance Has a b => Has a (b @ c) where
  get (b :@: _) = get b

instance {-# OVERLAPS #-} Has b (a @ b) where
  get (_ :@: b) = b

instance (Eq a, Eq b) => Eq (a @ b) where
  (a1 :@: b1) == (a2 :@: b2) = a1 == a2 && b1 == b2

instance (Ord a, Ord b) => Ord (a @ b) where
  (a1 :@: b1) <= (a2 :@: b2) = a1 < a2 || (a1 == a2 && b1 <= b2)

instance (DefaultOrdered a, DefaultOrdered b) => DefaultOrdered (a @ b) where
  headerOrder _ = headerOrder (undefined :: a) <> headerOrder (undefined :: b)

instance (ToNamedRecord a, ToNamedRecord b) => ToNamedRecord (a @ b) where
  toNamedRecord (a :@: b) = Hash.union (toNamedRecord a) (toNamedRecord b)

instance (FromNamedRecord a, FromNamedRecord b) => FromNamedRecord (a @ b) where
  parseNamedRecord nr = (:@:) <$> parseNamedRecord nr <*> parseNamedRecord nr

class ToJSONObject a where
  toJSONObject :: a -> Object
  toJSONPairs :: a -> Series

instance {-# OVERLAPPABLE #-} (ToJSONObject a, ToJSONObject b) =>
  ToJSON (a @ b) where
  toJSON (a :@: b) = Object $ toJSONObject a `Object.union` toJSONObject b
  toEncoding (a :@: b) = pairs $ toJSONPairs a <> toJSONPairs b

instance {-# OVERLAPPABLE #-} (FromJSON a, FromJSON b) => FromJSON (a @ b) where
  parseJSON o = (:@:) <$> parseJSON o <*> parseJSON o
